import unittest
from textwrap import dedent, indent

import pytest

from rope.base.oi.type_hinting import evaluate
from rope.contrib.codeassist import code_assist
from ropetest import testutils


class AbstractHintingTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _assist(self, code, offset=None, resource=None, **kwds):
        if offset is None:
            offset = len(code)
        return code_assist(self.project, code, offset, resource, **kwds)

    def assert_completion_in_result(self, name, scope, result):
        for proposal in result:
            if proposal.name == name and proposal.scope == scope:
                return
        self.fail(
            "completion <%s> in scope %r not proposed, available names: %r"
            % (name, scope, [(i.name, i.scope) for i in result])
        )

    def assert_completion_not_in_result(self, name, scope, result):
        for proposal in result:
            if proposal.name == name and proposal.scope == scope:
                self.fail("completion <%s> was proposed" % name)

    def run(self, result=None):
        if self.__class__.__name__.startswith("Abstract"):
            return
        super().run(result)


class DocstringParamHintingTest(AbstractHintingTest):
    def test_hint_param(self):
        code = dedent('''\
            class Sample(object):
                def a_method(self, a_arg):
                    """:type a_arg: threading.Thread"""
                    a_arg.is_a''')
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hierarchical_hint_param(self):
        code = dedent('''\
            class ISample(object):
                def a_method(self, a_arg):
                    """:type a_arg: threading.Thread"""


            class Sample(ISample):
                def a_method(self, a_arg):
                    a_arg.is_a''')
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)


class DocstringReturnHintingTest(AbstractHintingTest):
    def test_hierarchical_hint_rtype(self):
        code = dedent('''\
            class ISample(object):
                def b_method(self):
                    """:rtype: threading.Thread"""


            class Sample(ISample):
                def b_method(self):
                    pass
                def a_method(self):
                    self.b_method().is_a''')
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)


def fix_indents(hint):
    return indent(hint, " " * 12).strip()


class AbstractAssignmentHintingTest(AbstractHintingTest):
    def _make_class_hint(self, type_str):
        raise NotImplementedError

    def _make_constructor_hint(self, type_str):
        raise NotImplementedError

    def test_hint_attr(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("threading.Thread"))}
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hierarchical_hint_attr(self):
        code = dedent(f"""\
            class ISample(object):
                {fix_indents(self._make_class_hint("threading.Thread"))}


            class Sample(ISample):
                a_attr = None
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_defined_by_constructor(self):
        code = dedent(f"""\
            class Sample(object):
                def __init__(self, arg):
                    {fix_indents(self._make_constructor_hint("threading.Thread"))}
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_attr_redefined_by_constructor(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("threading.Thread"))}
                def __init__(self):
                    self.a_attr = None
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hierarchical_hint_attr_redefined_by_constructor(self):
        code = dedent(f"""\
            class ISample(object):
                {fix_indents(self._make_class_hint("threading.Thread"))}


            class Sample(ISample):
                def __init__(self):
                    self.a_attr = None
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_attr_for_pre_defined_type(self):
        code = dedent(f"""\
            class Other(object):
                def is_alive(self):
                    pass


            class Sample(object):
                {fix_indents(self._make_class_hint("Other"))}
                def a_method(self):
                    self.a_attr.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_attr_for_post_defined_type(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("Other"))}
                def a_method(self):
                    self.a_attr.is_a""")
        offset = len(code)
        # Note: the leading blank lines are required.
        code += dedent("""\


            class Other(object):
                def is_alive(self):
                    pass
        """)
        result = self._assist(code, offset)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_list(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("list[threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr:
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_tuple(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("tuple[threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr:
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_set(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("set[threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr:
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_iterable(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("collections.Iterable[threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr:
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_iterator(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("collections.Iterator[threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr:
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_dict_key(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("dict[str, threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr.keys():
                        i.sta""")
        result = self._assist(code)
        self.assert_completion_in_result("startswith", "builtin", result)

    def test_hint_parametrized_dict_value(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("dict[str, threading.Thread]"))}
                def a_method(self):
                    for i in self.a_attr.values():
                        i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_parametrized_nested_tuple_list(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("tuple[list[threading.Thread]]"))}
                def a_method(self):
                    for j in self.a_attr:
                        for i in j:
                            i.is_a""")
        result = self._assist(code)
        self.assert_completion_in_result("is_alive", "attribute", result)

    def test_hint_or(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("str | threading.Thread"))}
                def a_method(self):
                    for i in self.a_attr.values():
                        i.is_a""")
        result = self._assist(code)
        try:
            # Be sure, there isn't errors currently
            self.assert_completion_in_result('is_alive', 'attribute', result)
        except AssertionError as e:
            pytest.xfail("failing configuration (but should work)")

    def test_hint_nonexistent(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("sdfdsf.asdfasdf.sdfasdf.Dffg"))}
                def a_method(self):
                    for i in self.a_attr.values():
                        i.is_a""")
        result = self._assist(code)
        self.assertEqual(result, [])

    def test_hint_invalid_syntax(self):
        code = dedent(f"""\
            class Sample(object):
                {fix_indents(self._make_class_hint("sdf | & # &*"))}
                def a_method(self):
                    for i in self.a_attr.values():
                        i.is_a""")
        result = self._assist(code)
        self.assertEqual(result, [])


class DocstringNoneAssignmentHintingTest(AbstractAssignmentHintingTest):
    def _make_class_hint(self, type_str):
        hint = dedent(f'''\
            """:type a_attr: {type_str}"""
            a_attr = None
        ''')
        return indent(hint, " " * 4)

    def _make_constructor_hint(self, type_str):
        hint = dedent(f'''\
            """:type arg: {type_str}"""
            self.a_attr = arg
        ''')
        return indent(hint, " " * 8)


class DocstringNotImplementedAssignmentHintingTest(AbstractAssignmentHintingTest):
    def _make_class_hint(self, type_str):
        hint = dedent(f'''\
            """:type a_attr: {type_str}"""
            a_attr = NotImplemented
        ''')
        return indent(hint, " " * 4)

    def _make_constructor_hint(self, type_str):
        hint = dedent(f'''\
            """:type arg: {type_str}"""
            self.a_attr = arg
        ''')
        return indent(hint, " " * 8)


class PEP0484CommentNoneAssignmentHintingTest(AbstractAssignmentHintingTest):
    def _make_class_hint(self, type_str):
        hint = dedent(f"""\
            a_attr = None  # type: {type_str}
        """)
        return indent(hint, " " * 4)

    def _make_constructor_hint(self, type_str):
        hint = dedent(f"""\
            self.a_attr = None  # type: {type_str}
        """)
        return indent(hint, " " * 8)


class PEP0484CommentNotImplementedAssignmentHintingTest(AbstractAssignmentHintingTest):
    def _make_class_hint(self, type_str):
        hint = dedent(f"""\
            a_attr = NotImplemented  # type: {type_str}
        """)
        return indent(hint, " " * 4)

    def _make_constructor_hint(self, type_str):
        hint = dedent(f"""\
            self.a_attr = NotImplemented  # type: {type_str}
        """)
        return indent(hint, " " * 8)


class EvaluateTest(unittest.TestCase):
    def test_parser(self):
        tests = [
            ("Foo", "(name Foo)"),
            ("mod1.Foo", "(name mod1.Foo)"),
            ("mod1.mod2.Foo", "(name mod1.mod2.Foo)"),
            ("Foo[Bar]", "('[' (name Foo) [(name Bar)])"),
            (
                "Foo[Bar1, Bar2, Bar3]",
                "('[' (name Foo) [(name Bar1), (name Bar2), (name Bar3)])",
            ),
            ("Foo[Bar[Baz]]", "('[' (name Foo) [('[' (name Bar) [(name Baz)])])"),
            (
                "Foo[Bar1[Baz1], Bar2[Baz2]]",
                "('[' (name Foo) [('[' (name Bar1) [(name Baz1)]), ('[' (name Bar2) [(name Baz2)])])",
            ),
            ("mod1.mod2.Foo[Bar]", "('[' (name mod1.mod2.Foo) [(name Bar)])"),
            (
                "mod1.mod2.Foo[mod1.mod2.Bar]",
                "('[' (name mod1.mod2.Foo) [(name mod1.mod2.Bar)])",
            ),
            (
                "mod1.mod2.Foo[Bar1, Bar2, Bar3]",
                "('[' (name mod1.mod2.Foo) [(name Bar1), (name Bar2), (name Bar3)])",
            ),
            (
                "mod1.mod2.Foo[mod1.mod2.Bar[mod1.mod2.Baz]]",
                "('[' (name mod1.mod2.Foo) [('[' (name mod1.mod2.Bar) [(name mod1.mod2.Baz)])])",
            ),
            (
                "mod1.mod2.Foo[mod1.mod2.Bar1[mod1.mod2.Baz1], mod1.mod2.Bar2[mod1.mod2.Baz2]]",
                "('[' (name mod1.mod2.Foo) [('[' (name mod1.mod2.Bar1) [(name mod1.mod2.Baz1)]), ('[' (name mod1.mod2.Bar2) [(name mod1.mod2.Baz2)])])",
            ),
            ("(Foo, Bar) -> Baz", "('(' [(name Foo), (name Bar)] (name Baz))"),
            (
                "(mod1.mod2.Foo[mod1.mod2.Bar1[mod1.mod2.Baz1], mod1.mod2.Bar2[mod1.mod2.Baz2]], mod1.mod2.Bar[mod1.mod2.Bar1[mod1.mod2.Baz1], mod1.mod2.Bar2[mod1.mod2.Baz2]]) -> mod1.mod2.Baz[mod1.mod2.Bar1[mod1.mod2.Baz1], mod1.mod2.Bar2[mod1.mod2.Baz2]]",
                "('(' [('[' (name mod1.mod2.Foo) [('[' (name mod1.mod2.Bar1) [(name mod1.mod2.Baz1)]), ('[' (name mod1.mod2.Bar2) [(name mod1.mod2.Baz2)])]), ('[' (name mod1.mod2.Bar) [('[' (name mod1.mod2.Bar1) [(name mod1.mod2.Baz1)]), ('[' (name mod1.mod2.Bar2) [(name mod1.mod2.Baz2)])])] ('[' (name mod1.mod2.Baz) [('[' (name mod1.mod2.Bar1) [(name mod1.mod2.Baz1)]), ('[' (name mod1.mod2.Bar2) [(name mod1.mod2.Baz2)])]))",
            ),
            (
                "(Foo, Bar) -> Baz | Foo[Bar[Baz]]",
                "('|' ('(' [(name Foo), (name Bar)] (name Baz)) ('[' (name Foo) [('[' (name Bar) [(name Baz)])]))",
            ),
            (
                "Foo[Bar[Baz | (Foo, Bar) -> Baz]]",
                "('[' (name Foo) [('[' (name Bar) [('|' (name Baz) ('(' [(name Foo), (name Bar)] (name Baz)))])])",
            ),
        ]

        for t, expected in tests:
            result = repr(evaluate.compile(t))
            self.assertEqual(expected, result)


class RegressionHintingTest(AbstractHintingTest):
    def test_hierarchical_hint_for_mutable_attr_type(self):
        """Test for #157, AttributeError: 'PyObject' object has no attribute 'get_doc'"""
        code = dedent("""\
            class SuperClass(object):
                def __init__(self):
                    self.foo = None


            class SubClass(SuperClass):
                def __init__(self):
                    super(SubClass, self).__init__()
                    self.bar = 3


                def foo(self):
                    return self.bar""")
        result = self._assist(code)
        self.assert_completion_in_result("bar", "attribute", result)



import unittest
from textwrap import dedent

import rope.base.libutils
import rope.base.oi
from rope.base.builtins import Str
from ropetest import testutils


class DynamicOITest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project(validate_objectdb=True)
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_dti(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                return eval("arg")
            a_var = a_func(a_func)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        self.assertEqual(pymod["a_func"].get_object(), pymod["a_var"].get_object())

    def test_module_dti(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        code = dedent("""\
            import mod1
            def a_func(arg):
                return eval("arg")
            a_var = a_func(mod1)
        """)
        mod2.write(code)
        self.pycore.run_module(mod2).wait_process()
        pymod2 = self.project.get_pymodule(mod2)
        self.assertEqual(self.project.get_pymodule(mod1), pymod2["a_var"].get_object())

    def test_class_from_another_module_dti(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        code1 = dedent("""\
            class AClass(object):
                pass
        """)
        code2 = dedent("""\
            from mod1 import AClass

            def a_func(arg):
                return eval("arg")
            a_var = a_func(AClass)
        """)
        mod1.write(code1)
        mod2.write(code2)
        self.pycore.run_module(mod2).wait_process()
        # pymod1 = self.project.get_pymodule(mod1)
        pymod2 = self.project.get_pymodule(mod2)
        self.assertEqual(pymod2["AClass"].get_object(), pymod2["a_var"].get_object())

    def test_class_dti(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class AClass(object):
                pass

            def a_func(arg):
                return eval("arg")
            a_var = a_func(AClass)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        self.assertEqual(pymod["AClass"].get_object(), pymod["a_var"].get_object())

    def test_instance_dti(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class AClass(object):
                pass

            def a_func(arg):
                return eval("arg()")
            a_var = a_func(AClass)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        self.assertEqual(
            pymod["AClass"].get_object(), pymod["a_var"].get_object().get_type()
        )

    def test_method_dti(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class AClass(object):
                def a_method(self, arg):
                    return eval("arg()")
            an_instance = AClass()
            a_var = an_instance.a_method(AClass)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        self.assertEqual(
            pymod["AClass"].get_object(), pymod["a_var"].get_object().get_type()
        )

    def test_function_argument_dti(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                pass
            a_func(a_func)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pyscope = self.project.get_pymodule(mod).get_scope()
        self.assertEqual(
            pyscope["a_func"].get_object(), pyscope.get_scopes()[0]["arg"].get_object()
        )

    def test_classes_with_the_same_name(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                class AClass(object):
                    pass
                return eval("arg")
            class AClass(object):
                pass
            a_var = a_func(AClass)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        self.assertEqual(pymod["AClass"].get_object(), pymod["a_var"].get_object())

    def test_nested_classes(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func():
                class AClass(object):
                    pass
                return AClass
            def another_func(arg):
                return eval("arg")
            a_var = another_func(a_func())
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pyscope = self.project.get_pymodule(mod).get_scope()
        self.assertEqual(
            pyscope.get_scopes()[0]["AClass"].get_object(),
            pyscope["a_var"].get_object(),
        )

    def test_function_argument_dti2(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg, a_builtin_type):
                pass
            a_func(a_func, [])
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pyscope = self.project.get_pymodule(mod).get_scope()
        self.assertEqual(
            pyscope["a_func"].get_object(), pyscope.get_scopes()[0]["arg"].get_object()
        )

    def test_dti_and_concluded_data_invalidation(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                return eval("arg")
            a_var = a_func(a_func)
        """)
        mod.write(code)
        pymod = self.project.get_pymodule(mod)
        pymod["a_var"].get_object()
        self.pycore.run_module(mod).wait_process()
        self.assertEqual(pymod["a_func"].get_object(), pymod["a_var"].get_object())

    def test_list_objects_and_dynamicoi(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                pass
            def a_func(arg):
                return eval("arg")
            a_var = a_func([C()])[0]
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_for_loops_and_dynamicoi(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                pass
            def a_func(arg):
                return eval("arg")
            for c in a_func([C()]):
                a_var = c
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_dict_objects_and_dynamicoi(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                pass
            def a_func(arg):
                return eval("arg")
            a_var = a_func({1: C()})[1]
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_dict_keys_and_dynamicoi(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
                class C(object):
                    pass
                def a_func(arg):
                    return eval("arg")
                a_var = list(a_func({C(): 1}))[0]
            """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_dict_keys_and_dynamicoi2(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(arg):
                return eval("arg")
            a, b = a_func((C1(), C2()))
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_strs_and_dynamicoi(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                return eval("arg")
            a_var = a_func("hey")
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), rope.base.builtins.Str))

    def test_textual_transformations(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                pass
            def f():
                pass
            a_var = C()
            a_list = [C()]
            a_str = "hey"
            a_file = open("file.txt")
        """)
        mod.write(code)
        to_pyobject = rope.base.oi.transform.TextualToPyObject(self.project)
        to_textual = rope.base.oi.transform.PyObjectToTextual(self.project)
        pymod = self.project.get_pymodule(mod)

        def complex_to_textual(pyobject):
            return to_textual.transform(
                to_pyobject.transform(to_textual.transform(pyobject))
            )

        test_variables = [
            ("C", ("defined", "mod.py", "C")),
            ("f", ("defined", "mod.py", "f")),
            ("a_var", ("instance", ("defined", "mod.py", "C"))),
            ("a_list", ("builtin", "list", ("instance", ("defined", "mod.py", "C")))),
            ("a_str", ("builtin", "str")),
            ("a_file", ("builtin", "file")),
        ]
        test_cases = [(pymod[v].get_object(), r) for v, r in test_variables]
        test_cases += [
            (pymod, ("defined", "mod.py")),
            (
                rope.base.builtins.builtins["enumerate"].get_object(),
                ("builtin", "function", "enumerate"),
            ),
        ]
        for var, result in test_cases:
            self.assertEqual(to_textual.transform(var), result)
            self.assertEqual(complex_to_textual(var), result)

    def test_arguments_with_keywords(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(arg):
                return eval("arg")
            a = a_func(arg=C1())
            b = a_func(arg=C2())
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_a_function_with_different_returns(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(arg):
                return eval("arg")
            a = a_func(C1())
            b = a_func(C2())
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_a_function_with_different_returns2(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(p):
                if p == C1:
                    return C1()
                else:
                    return C2()
            a = a_func(C1)
            b = a_func(C2)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_ignoring_star_args(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(p, *args):
                if p == C1:
                    return C1()
                else:
                    return C2()
            a = a_func(C1, 1)
            b = a_func(C2, 2)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_ignoring_double_star_args(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(p, *kwds, **args):
                if p == C1:
                    return C1()
                else:
                    return C2()
            a = a_func(C1, kwd=1)
            b = a_func(C2, kwd=2)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        pymod = self.project.get_pymodule(mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_invalidating_data_after_changing(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            def a_func(arg):
                return eval("arg")
            a_var = a_func(a_func)
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        mod.write(code.replace("a_func", "newfunc"))
        mod.write(code)
        pymod = self.project.get_pymodule(mod)
        self.assertNotEqual(pymod["a_func"].get_object(), pymod["a_var"].get_object())

    def test_invalidating_data_after_moving(self):
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write("class C(object):\n    pass\n")
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            import mod2
            def a_func(arg):
                return eval(arg)
            a_var = a_func("mod2.C")
        """)
        mod.write(code)
        self.pycore.run_module(mod).wait_process()
        mod.move("newmod.py")
        pymod = self.project.get_module("newmod")
        pymod2 = self.project.get_pymodule(mod2)
        self.assertEqual(pymod2["C"].get_object(), pymod["a_var"].get_object())


class NewStaticOITest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project(validate_objectdb=True)
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_static_oi_for_simple_function_calls(self):
        code = dedent("""\
            class C(object):
                pass
            def f(p):
                pass
            f(C())
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        f_scope = pymod["f"].get_object().get_scope()
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(c_class, p_type)

    def test_static_oi_not_failing_when_callin_callables(self):
        code = dedent("""\
            class C(object):
                pass
            C()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)

    def test_static_oi_for_nested_calls(self):
        code = dedent("""\
            class C(object):
                pass
            def f(p):
                pass
            def g(p):
                return p
            f(g(C()))
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        f_scope = pymod["f"].get_object().get_scope()
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(c_class, p_type)

    def test_static_oi_class_methods(self):
        code = dedent("""\
            class C(object):
                def f(self, p):
                    pass
            C().f(C())""")
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        f_scope = c_class["f"].get_object().get_scope()
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(c_class, p_type)

    def test_static_oi_preventing_soi_maximum_recursion_exceptions(self):
        code = dedent("""\
            item = {}
            for item in item.keys():
                pass
        """)
        self.mod.write(code)
        try:
            self.pycore.analyze_module(self.mod)
        except RuntimeError as e:
            self.fail(str(e))

    def test_static_oi_for_infer_return_typs_from_funcs_based_on_params(self):
        code = dedent("""\
            class C(object):
                pass
            def func(p):
                return p
            a_var = func(C())
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_a_function_with_different_returns(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            def a_func(arg):
                return arg
            a = a_func(C1())
            b = a_func(C2())
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_not_reporting_out_of_date_information(self):
        code = dedent("""\
            class C1(object):
                pass
            def f(arg):
                return C1()
            a_var = f()
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c1_class, a_var.get_type())

        self.mod.write(code.replace("C1", "C2"))
        pymod = self.project.get_pymodule(self.mod)
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c2_class, a_var.get_type())

    def test_invalidating_concluded_data_in_a_function(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            def func(arg):
                temp = arg
                return temp
        """))
        mod2.write(dedent("""\
            import mod1
            class C1(object):
                pass
            class C2(object):
                pass
            a_var = mod1.func(C1())
        """))
        pymod2 = self.project.get_pymodule(mod2)
        c1_class = pymod2["C1"].get_object()
        a_var = pymod2["a_var"].get_object()
        self.assertEqual(c1_class, a_var.get_type())

        mod2.write(mod2.read()[: mod2.read().rfind("C1()")] + "C2())\n")
        pymod2 = self.project.get_pymodule(mod2)
        c2_class = pymod2["C2"].get_object()
        a_var = pymod2["a_var"].get_object()
        self.assertEqual(c2_class, a_var.get_type())

    def test_handling_generator_functions_for_strs(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            def f(p):
                yield p()
            for c in f(C):
                a_var = c
        """))
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    # TODO: Returning a generator for functions that yield unknowns
    @unittest.skip("Returning a generator that yields unknowns")
    def xxx_test_handl_generator_functions_when_unknown_type_is_yielded(self):
        self.mod.write(dedent("""\
            class C(object):
                pass
            def f():
                yield eval("C()")
            a_var = f()
        """))
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertTrue(isinstance(a_var.get_type(), rope.base.builtins.Generator))

    def test_static_oi_for_lists_depending_on_append_function(self):
        code = dedent("""\
            class C(object):
                pass
            l = list()
            l.append(C())
            a_var = l.pop()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_lists_per_object_for_get_item(self):
        code = dedent("""\
            class C(object):
                pass
            l = list()
            l.append(C())
            a_var = l[0]
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_lists_per_object_for_fields(self):
        code = dedent("""\
            class C(object):
                pass
            class A(object):
                def __init__(self):
                    self.l = []
                def set(self):
                    self.l.append(C())
            a = A()
            a.set()
            a_var = a.l[0]
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_lists_per_object_for_set_item(self):
        code = dedent("""\
            class C(object):
                pass
            l = [None]
            l[0] = C()
            a_var = l[0]
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_lists_per_object_for_extending_lists(self):
        code = dedent("""\
            class C(object):
                pass
            l = []
            l.append(C())
            l2 = []
            l2.extend(l)
            a_var = l2[0]
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_lists_per_object_for_iters(self):
        code = dedent("""\
            class C(object):
                pass
            l = []
            l.append(C())
            for c in l:
                a_var = c
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_static_oi_for_dicts_depending_on_append_function(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {}
            d[C1()] = C2()
            a, b = d.popitem()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_static_oi_for_dicts_depending_on_for_loops(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {}
            d[C1()] = C2()
            for k, v in d.items():
                a = k
                b = v
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_static_oi_for_dicts_depending_on_update(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {}
            d[C1()] = C2()
            d2 = {}
            d2.update(d)
            a, b = d2.popitem()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_static_oi_for_dicts_depending_on_update_on_seqs(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                pass
            d = {}
            d.update([(C1(), C2())])
            a, b = d.popitem()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        c2_class = pymod["C2"].get_object()
        a_var = pymod["a"].get_object()
        b_var = pymod["b"].get_object()
        self.assertEqual(c1_class, a_var.get_type())
        self.assertEqual(c2_class, b_var.get_type())

    def test_static_oi_for_sets_per_object_for_set_item(self):
        code = dedent("""\
            class C(object):
                pass
            s = set()
            s.add(C())
            a_var = s.pop()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c_class, a_var.get_type())

    def test_properties_and_calling_get_property(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                c1 = C1()
                def get_c1(self):
                    return self.c1
                p = property(get_c1)
            c2 = C2()
            a_var = c2.p
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c1_class, a_var.get_type())

    def test_soi_on_constructors(self):
        code = dedent("""\
            class C1(object):
                pass
            class C2(object):
                def __init__(self, arg):
                    self.attr = arg
            c2 = C2(C1())
            a_var = c2.attr""")
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        a_var = pymod["a_var"].get_object()
        self.assertEqual(c1_class, a_var.get_type())

    def test_soi_on_literal_assignment(self):
        code = 'a_var = ""'
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertEqual(Str, type(a_var.get_type()))

    @testutils.only_for_versions_higher("3.6")
    def test_soi_on_typed_assignment(self):
        code = "a_var: str"
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        a_var = pymod["a_var"].get_object()
        self.assertEqual(Str, type(a_var.get_type()))

    def test_not_saving_unknown_function_returns(self):
        mod2 = testutils.create_module(self.project, "mod2")
        self.mod.write(dedent("""\
            class C(object):
                pass
            l = []
            l.append(C())
        """))
        mod2.write(dedent("""\
            import mod
            def f():
                return mod.l.pop()
            a_var = f()
        """))
        pymod = self.project.get_pymodule(self.mod)
        pymod2 = self.project.get_pymodule(mod2)
        c_class = pymod["C"].get_object()
        a_var = pymod2["a_var"]

        self.pycore.analyze_module(mod2)
        self.assertNotEqual(c_class, a_var.get_object().get_type())

        self.pycore.analyze_module(self.mod)
        self.assertEqual(c_class, a_var.get_object().get_type())

    def test_using_the_best_callinfo(self):
        code = dedent("""\
            class C1(object):
                pass
            def f(arg1, arg2, arg3):
                pass
            f("", None, C1())
            f("", C1(), None)
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        c1_class = pymod["C1"].get_object()
        f_scope = pymod["f"].get_object().get_scope()
        arg2 = f_scope["arg2"].get_object()
        self.assertEqual(c1_class, arg2.get_type())

    def test_call_function_and_parameters(self):
        code = dedent("""\
            class A(object):
                def __call__(self, p):
                    pass
            A()("")
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        scope = self.project.get_pymodule(self.mod).get_scope()
        p_object = scope.get_scopes()[0].get_scopes()[0]["p"].get_object()
        self.assertTrue(isinstance(p_object.get_type(), rope.base.builtins.Str))

    def test_report_change_in_libutils(self):
        self.project.prefs["automatic_soa"] = True
        code = dedent("""\
            class C(object):
                pass
            def f(p):
                pass
            f(C())
        """)
        with open(self.mod.real_path, "w") as mod_file:
            mod_file.write(code)

        rope.base.libutils.report_change(self.project, self.mod.real_path, "")
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        f_scope = pymod["f"].get_object().get_scope()
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(c_class, p_type)

    def test_report_libutils_and_analyze_all_modules(self):
        code = dedent("""\
            class C(object):
                pass
            def f(p):
                pass
            f(C())
        """)
        self.mod.write(code)
        rope.base.libutils.analyze_modules(self.project)
        pymod = self.project.get_pymodule(self.mod)
        c_class = pymod["C"].get_object()
        f_scope = pymod["f"].get_object().get_scope()
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(c_class, p_type)

    def test_validation_problems_for_objectdb_retrievals(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write(dedent("""\
            l = []
            var = l.pop()
        """))
        mod2.write(dedent("""\
            import mod1

            class C(object):
                pass
            mod1.l.append(C())
        """))
        self.pycore.analyze_module(mod2)

        pymod2 = self.project.get_pymodule(mod2)
        c_class = pymod2["C"].get_object()
        pymod1 = self.project.get_pymodule(mod1)
        var_pyname = pymod1["var"]
        self.assertEqual(c_class, var_pyname.get_object().get_type())
        mod2.write(dedent("""\
            import mod1

            mod1.l.append("")
        """))
        self.assertNotEqual(
            c_class, var_pyname.get_object().get_type(), "Class `C` no more exists"
        )

    def test_validation_problems_for_changing_builtin_types(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            l = []
            l.append("")
        """))
        self.pycore.analyze_module(mod1)

        mod1.write(dedent("""\
            l = {}
            v = l["key"]
        """))
        pymod1 = self.project.get_pymodule(mod1)
        var = pymod1["v"].get_object()  # noqa

    def test_always_returning_containing_class_for_selfs(self):
        code = dedent("""\
            class A(object):
                def f(p):
                    return p
            class B(object):
                pass
            b = B()
            b.f()
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod)
        pymod = self.project.get_pymodule(self.mod)
        a_class = pymod["A"].get_object()
        f_scope = a_class.get_scope().get_scopes()[0]
        p_type = f_scope["p"].get_object().get_type()
        self.assertEqual(a_class, p_type)

    def test_following_function_calls_when_asked_to(self):
        code = dedent("""\
            class A(object):
                pass
            class C(object):
                def __init__(self, arg):
                    self.attr = arg
            def f(p):
                return C(p)
            c = f(A())
            x = c.attr
        """)
        self.mod.write(code)
        self.pycore.analyze_module(self.mod, followed_calls=1)
        pymod = self.project.get_pymodule(self.mod)
        a_class = pymod["A"].get_object()
        x_var = pymod["x"].get_object().get_type()
        self.assertEqual(a_class, x_var)

    def test_set_comprehension(self):
        code = dedent("""\
            x = {s.strip() for s in X()}
            x.add('x')
        """)
        self.mod.write(code)
        pymod = self.project.get_pymodule(self.mod)
        x_var = pymod["x"].pyobject.get()



from rope.base import codeanalyze, evaluate, exceptions, libutils, utils, worder
from rope.base.codeanalyze import ArrayLinesAdapter, LogicalLineFinder


class FixSyntax:
    def __init__(self, project, code, resource, maxfixes=1):
        self.project = project
        self.code = code
        self.resource = resource
        self.maxfixes = maxfixes

    @utils.saveit
    def get_pymodule(self):
        """Get a `PyModule`"""
        msg = None
        code = self.code
        tries = 0
        while True:
            try:
                if (
                    tries == 0
                    and self.resource is not None
                    and self.resource.read() == code
                ):
                    return self.project.get_pymodule(self.resource, force_errors=True)
                return libutils.get_string_module(
                    self.project, code, resource=self.resource, force_errors=True
                )
            except exceptions.ModuleSyntaxError as e:
                if msg is None:
                    msg = f"{e.filename}:{e.lineno} {e.message_}"
                if tries < self.maxfixes:
                    tries += 1
                    self.commenter.comment(e.lineno)
                    code = "\n".join(self.commenter.lines)
                else:
                    raise exceptions.ModuleSyntaxError(
                        e.filename, e.lineno, f"Failed to fix error: {msg}"
                    )

    @property
    @utils.saveit
    def commenter(self):
        return _Commenter(self.code)

    def pyname_at(self, offset):
        pymodule = self.get_pymodule()

        def old_pyname():
            word_finder = worder.Worder(self.code, True)
            expression = word_finder.get_primary_at(offset)
            expression = expression.replace("\\\n", " ").replace("\n", " ")
            lineno = self.code.count("\n", 0, offset)
            scope = pymodule.get_scope().get_inner_scope_for_line(lineno)
            return evaluate.eval_str(scope, expression)

        new_code = pymodule.source_code

        def new_pyname():
            newoffset = self.commenter.transferred_offset(offset)
            return evaluate.eval_location(pymodule, newoffset)

        if new_code.startswith(self.code[: offset + 1]):
            return new_pyname()
        result = old_pyname()
        if result is None:
            return new_pyname()
        return result


class _Commenter:
    def __init__(self, code):
        self.code = code
        self.lines = self.code.split("\n")
        self.lines.append("\n")
        self.origs = list(range(len(self.lines) + 1))
        self.diffs = [0] * (len(self.lines) + 1)

    def comment(self, lineno):
        start = _logical_start(self.lines, lineno, check_prev=True) - 1
        # using self._get_stmt_end() instead of self._get_block_end()
        # to lower commented lines
        end = self._get_stmt_end(start)
        indents = _get_line_indents(self.lines[start])
        if 0 < start:
            last_lineno = self._last_non_blank(start - 1)
            last_line = self.lines[last_lineno]
            if last_line.rstrip().endswith(":"):
                indents = _get_line_indents(last_line) + 4
        self._set(start, " " * indents + "pass")
        for line in range(start + 1, end + 1):
            self._set(line, self.lines[start])
        self._fix_incomplete_try_blocks(lineno, indents)

    def transferred_offset(self, offset):
        lineno = self.code.count("\n", 0, offset)
        diff = sum(self.diffs[:lineno])
        return offset + diff

    def _last_non_blank(self, start):
        while start > 0 and self.lines[start].strip() == "":
            start -= 1
        return start

    def _get_block_end(self, lineno):
        end_line = lineno
        base_indents = _get_line_indents(self.lines[lineno])
        for i in range(lineno + 1, len(self.lines)):
            if _get_line_indents(self.lines[i]) >= base_indents:
                end_line = i
            else:
                break
        return end_line

    def _get_stmt_end(self, lineno):
        base_indents = _get_line_indents(self.lines[lineno])
        for i in range(lineno + 1, len(self.lines)):
            if _get_line_indents(self.lines[i]) <= base_indents:
                return i - 1
        return lineno

    def _fix_incomplete_try_blocks(self, lineno, indents):
        block_start = lineno
        last_indents = indents
        while block_start > 0:
            block_start = (
                codeanalyze.get_block_start(ArrayLinesAdapter(self.lines), block_start)
                - 1
            )
            if self.lines[block_start].strip().startswith("try:"):
                indents = _get_line_indents(self.lines[block_start])
                if indents > last_indents:
                    continue
                last_indents = indents
                block_end = self._find_matching_deindent(block_start)
                line = self.lines[block_end].strip()
                if not (
                    line.startswith("finally:")
                    or line.startswith("except ")
                    or line.startswith("except:")
                ):
                    self._insert(block_end, " " * indents + "finally:")
                    self._insert(block_end + 1, " " * indents + "    pass")

    def _find_matching_deindent(self, line_number):
        indents = _get_line_indents(self.lines[line_number])
        current_line = line_number + 1
        while current_line < len(self.lines):
            line = self.lines[current_line]
            if not line.strip().startswith("#") and not line.strip() == "":
                # HACK: We should have used logical lines here
                if _get_line_indents(self.lines[current_line]) <= indents:
                    return current_line
            current_line += 1
        return len(self.lines) - 1

    def _set(self, lineno, line):
        self.diffs[self.origs[lineno]] += len(line) - len(self.lines[lineno])
        self.lines[lineno] = line

    def _insert(self, lineno, line):
        self.diffs[self.origs[lineno]] += len(line) + 1
        self.origs.insert(lineno, self.origs[lineno])
        self.lines.insert(lineno, line)


def _logical_start(lines, lineno, check_prev=False):
    logical_finder = LogicalLineFinder(ArrayLinesAdapter(lines))
    if check_prev:
        prev = lineno - 1
        while prev > 0:
            start, end = logical_finder.logical_line_in(prev)
            if end is None or start <= lineno < end:
                return start
            if start <= prev:
                break
            prev -= 1
    return logical_finder.logical_line_in(lineno)[0]


def _get_line_indents(line):
    return codeanalyze.count_line_indents(line)



import unittest
from textwrap import dedent

from rope.base import exceptions
from rope.refactor.usefunction import UseFunction
from ropetest import testutils


class UseFunctionTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.mod1 = testutils.create_module(self.project, "mod1")
        self.mod2 = testutils.create_module(self.project, "mod2")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_case(self):
        code = dedent("""\
            def f():
                pass
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(code, self.mod1.read())

    def test_simple_function(self):
        code = dedent("""\
            def f(p):
                print(p)
            print(1)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    print(p)
                f(1)
            """),
            self.mod1.read(),
        )

    def test_simple_function2(self):
        code = dedent("""\
            def f(p):
                print(p + 1)
            print(1 + 1)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    print(p + 1)
                f(1)
            """),
            self.mod1.read(),
        )

    def test_functions_with_multiple_statements(self):
        code = dedent("""\
            def f(p):
                r = p + 1
                print(r)
            r = 2 + 1
            print(r)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    r = p + 1
                    print(r)
                f(2)
            """),
            self.mod1.read(),
        )

    def test_returning(self):
        code = dedent("""\
            def f(p):
                return p + 1
            r = 2 + 1
            print(r)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    return p + 1
                r = f(2)
                print(r)
            """),
            self.mod1.read(),
        )

    def test_returning_a_single_expression(self):
        code = dedent("""\
            def f(p):
                return p + 1
            print(2 + 1)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    return p + 1
                print(f(2))
            """),
            self.mod1.read(),
        )

    def test_occurrences_in_other_modules(self):
        code = dedent("""\
            def f(p):
                return p + 1
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.mod2.write("print(2 + 1)\n")
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                import mod1
                print(mod1.f(2))
            """),
            self.mod2.read(),
        )

    def test_when_performing_on_non_functions(self):
        code = "var = 1\n"
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            UseFunction(self.project, self.mod1, code.rindex("var"))

    def test_differing_in_the_inner_temp_names(self):
        code = dedent("""\
            def f(p):
                a = p + 1
                print(a)
            b = 2 + 1
            print(b)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    a = p + 1
                    print(a)
                f(2)
            """),
            self.mod1.read(),
        )

    # TODO: probably new options should be added to restructure
    def xxx_test_being_a_bit_more_intelligent_when_returning_assigneds(self):
        code = dedent("""\
            def f(p):
                a = p + 1
                return a
            var = 2 + 1
            print(var)
        """)
        self.mod1.write(code)
        user = UseFunction(self.project, self.mod1, code.rindex("f"))
        self.project.do(user.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    a = p + 1
                    return a
                var = f(p)
                print(var)
            """),
            self.mod1.read(),
        )

    def test_exception_when_performing_a_function_with_yield(self):
        code = dedent("""\
            def func():
                yield 1
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            UseFunction(self.project, self.mod1, code.index("func"))

    def test_exception_when_performing_a_function_two_returns(self):
        code = dedent("""\
            def func():
                return 1
                return 2
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            UseFunction(self.project, self.mod1, code.index("func"))

    def test_exception_when_returns_is_not_the_last_statement(self):
        code = dedent("""\
            def func():
                return 2
                a = 1
        """)
        self.mod1.write(code)
        with self.assertRaises(exceptions.RefactoringError):
            UseFunction(self.project, self.mod1, code.index("func"))



"""Definitions of types for the Autoimport program."""
import pathlib
from enum import Enum
from typing import NamedTuple, Optional


class Source(Enum):
    """Describes the source of the package, for sorting purposes."""

    PROJECT = 0  # Obviously any project packages come first
    MANUAL = 1  # Placeholder since Autoimport classifies manually added modules
    BUILTIN = 2
    STANDARD = 3  # We want to favor standard library items
    SITE_PACKAGE = 4
    UNKNOWN = 5

    # modified_time


class ModuleInfo(NamedTuple):
    """Descriptor of information to get names from a module."""

    filepath: Optional[pathlib.Path]
    modname: str
    underlined: bool
    process_imports: bool


class ModuleFile(ModuleInfo):
    """Descriptor of information to get names from a file using ast."""

    filepath: pathlib.Path
    modname: str
    underlined: bool
    process_imports: bool


class ModuleCompiled(ModuleInfo):
    """Descriptor of information to get names using imports."""

    filepath = None
    modname: str
    underlined: bool
    process_imports: bool


class PackageType(Enum):
    """Describes the type of package, to determine how to get the names from it."""

    BUILTIN = 0  # No file exists, compiled into python. IE: Sys
    STANDARD = 1  # Just a folder
    COMPILED = 2  # .so module
    SINGLE_FILE = 3  # a .py file


class NameType(Enum):
    """Describes the type of Name for lsp completions. Taken from python lsp server."""

    Text = 1
    Method = 2
    Function = 3
    Constructor = 4
    Field = 5
    Variable = 6
    Class = 7
    Interface = 8
    Module = 9
    Property = 10
    Unit = 11
    Value = 12
    Enum = 13
    Keyword = 14
    Snippet = 15
    Color = 16
    File = 17
    Reference = 18
    Folder = 19
    EnumMember = 20
    Constant = 21
    Struct = 22
    Event = 23
    Operator = 24
    TypeParameter = 25


class Package(NamedTuple):
    """Attributes of a package."""

    name: str
    source: Source
    path: Optional[pathlib.Path]
    type: PackageType


class Name(NamedTuple):
    """A Name to be added to the database."""

    name: str
    modname: str
    package: str
    source: Source
    name_type: NameType


class PartialName(NamedTuple):
    """Partial information of a Name."""

    name: str
    name_type: NameType


class SearchResult(NamedTuple):
    """Search Result."""

    import_statement: str
    name: str
    source: int
    itemkind: int



"""A module containing classes for move refactoring

`create_move()` is a factory for creating move refactoring objects
based on inputs.

"""
from __future__ import annotations

import typing
from typing import Union

from rope.base import (
    codeanalyze,
    evaluate,
    exceptions,
    libutils,
    pynames,
    pyobjects,
    taskhandle,
    worder,
)
from rope.base.change import ChangeContents, ChangeSet, MoveResource
from rope.refactor import functionutils, importutils, occurrences, rename, sourceutils
from rope.refactor.importutils.module_imports import (
    get_first_decorator_or_function_start_line,
)

if typing.TYPE_CHECKING:
    from rope.base import project, resources


def create_move(project, resource, offset=None):
    """A factory for creating Move objects

    Based on `resource` and `offset`, return one of `MoveModule`,
    `MoveGlobal` or `MoveMethod` for performing move refactoring.

    """
    if offset is None:
        return MoveModule(project, resource)
    this_pymodule = project.get_pymodule(resource)
    pyname = evaluate.eval_location(this_pymodule, offset)
    if pyname is not None:
        pyobject = pyname.get_object()
        if isinstance(pyobject, pyobjects.PyModule) or isinstance(
            pyobject, pyobjects.PyPackage
        ):
            return MoveModule(project, pyobject.get_resource())
        if isinstance(pyobject, pyobjects.PyFunction) and isinstance(
            pyobject.parent, pyobjects.PyClass
        ):
            return MoveMethod(project, resource, offset)
        if (
            isinstance(pyobject, pyobjects.PyDefinedObject)
            and isinstance(pyobject.parent, pyobjects.PyModule)
            or isinstance(pyname, pynames.AssignedName)
        ):
            return MoveGlobal(project, resource, offset)
    raise exceptions.RefactoringError(
        "Move only works on global classes/functions/variables, modules and methods."
    )


class MoveMethod:
    """For moving methods

    It makes a new method in the destination class and changes
    the body of the old method to call the new method.  You can
    inline the old method to change all of its occurrences.

    """

    def __init__(self, project, resource, offset):
        self.project = project
        this_pymodule = self.project.get_pymodule(resource)
        pyname = evaluate.eval_location(this_pymodule, offset)
        self.method_name = worder.get_name_at(resource, offset)
        self.pyfunction = pyname.get_object()
        if self.pyfunction.get_kind() != "method":
            raise exceptions.RefactoringError("Only normal methods can be moved.")

    def get_changes(
        self,
        dest_attr,
        new_name=None,
        resources=None,
        task_handle=taskhandle.DEFAULT_TASK_HANDLE,  # FIXME: this is unused
    ):
        """Return the changes needed for this refactoring

        Parameters:

        - `dest_attr`: the name of the destination attribute
        - `new_name`: the name of the new method; if `None` uses
          the old name
        - `resources` can be a list of `rope.base.resources.File` to
          apply this refactoring on.  If `None`, the restructuring
          will be applied to all python files.

        """
        changes = ChangeSet("Moving method <%s>" % self.method_name)
        if resources is None:
            resources = self.project.get_python_files()
        if new_name is None:
            new_name = self.get_method_name()
        resource1, start1, end1, new_content1 = self._get_changes_made_by_old_class(
            dest_attr, new_name
        )
        collector1 = codeanalyze.ChangeCollector(resource1.read())
        collector1.add_change(start1, end1, new_content1)

        resource2, start2, end2, new_content2 = self._get_changes_made_by_new_class(
            dest_attr, new_name
        )
        if resource1 == resource2:
            collector1.add_change(start2, end2, new_content2)
        else:
            collector2 = codeanalyze.ChangeCollector(resource2.read())
            collector2.add_change(start2, end2, new_content2)
            result = collector2.get_changed()
            import_tools = importutils.ImportTools(self.project)
            new_imports = self._get_used_imports(import_tools)
            if new_imports:
                goal_pymodule = libutils.get_string_module(
                    self.project, result, resource2
                )
                result = _add_imports_to_module(
                    import_tools, goal_pymodule, new_imports
                )
            if resource2 in resources:
                changes.add_change(ChangeContents(resource2, result))

        if resource1 in resources:
            changes.add_change(ChangeContents(resource1, collector1.get_changed()))
        return changes

    def get_method_name(self):
        return self.method_name

    def _get_used_imports(self, import_tools):
        return importutils.get_imports(self.project, self.pyfunction)

    def _get_changes_made_by_old_class(self, dest_attr, new_name):
        pymodule = self.pyfunction.get_module()
        indents = self._get_scope_indents(self.pyfunction)
        body = "return self.{}.{}({})\n".format(
            dest_attr,
            new_name,
            self._get_passed_arguments_string(),
        )
        region = sourceutils.get_body_region(self.pyfunction)
        return (
            pymodule.get_resource(),
            region[0],
            region[1],
            sourceutils.fix_indentation(body, indents),
        )

    def _get_scope_indents(self, pyobject):
        pymodule = pyobject.get_module()
        return sourceutils.get_indents(
            pymodule.lines, pyobject.get_scope().get_start()
        ) + sourceutils.get_indent(self.project)

    def _get_changes_made_by_new_class(self, dest_attr, new_name):
        old_pyclass = self.pyfunction.parent
        if dest_attr not in old_pyclass:
            raise exceptions.RefactoringError(
                "Destination attribute <%s> not found" % dest_attr
            )
        pyclass = old_pyclass[dest_attr].get_object().get_type()
        if not isinstance(pyclass, pyobjects.PyClass):
            raise exceptions.RefactoringError(
                "Unknown class type for attribute <%s>" % dest_attr
            )
        pymodule = pyclass.get_module()
        resource = pyclass.get_module().get_resource()
        start, end = sourceutils.get_body_region(pyclass)
        pre_blanks = "\n"
        if pymodule.source_code[start:end].strip() != "pass":
            pre_blanks = "\n\n"
            start = end
        indents = self._get_scope_indents(pyclass)
        body = pre_blanks + sourceutils.fix_indentation(
            self.get_new_method(new_name), indents
        )
        return resource, start, end, body

    def get_new_method(self, name):
        return "{}\n{}".format(
            self._get_new_header(name),
            sourceutils.fix_indentation(
                self._get_body(), sourceutils.get_indent(self.project)
            ),
        )

    def _get_unchanged_body(self):
        return sourceutils.get_body(self.pyfunction)

    def _get_body(self, host="host"):
        self_name = self._get_self_name()
        body = self_name + " = None\n" + self._get_unchanged_body()
        pymodule = libutils.get_string_module(self.project, body)
        finder = occurrences.create_finder(self.project, self_name, pymodule[self_name])
        result = rename.rename_in_module(finder, host, pymodule=pymodule)
        if result is None:
            result = body
        return result[result.index("\n") + 1 :]

    def _get_self_name(self):
        return self.pyfunction.get_param_names()[0]

    def _get_new_header(self, name):
        header = "def %s(self" % name
        if self._is_host_used():
            header += ", host"
        definition_info = functionutils.DefinitionInfo.read(self.pyfunction)
        others = definition_info.arguments_to_string(1)
        if others:
            header += ", " + others
        return header + "):"

    def _get_passed_arguments_string(self):
        result = ""
        if self._is_host_used():
            result = "self"
        definition_info = functionutils.DefinitionInfo.read(self.pyfunction)
        others = definition_info.arguments_to_string(1)
        if others:
            if result:
                result += ", "
            result += others
        return result

    def _is_host_used(self):
        return self._get_body("__old_self") != self._get_unchanged_body()


class MoveGlobal:
    """For moving global function and classes"""

    project: project.Project

    def __init__(self, project, resource, offset):
        self.project = project
        this_pymodule = self.project.get_pymodule(resource)
        self.old_pyname = evaluate.eval_location(this_pymodule, offset)
        if self.old_pyname is None:
            raise exceptions.RefactoringError(
                "Move refactoring should be performed on a class/function/variable."
            )
        if self._is_variable(self.old_pyname):
            self.old_name = worder.get_name_at(resource, offset)
            pymodule = this_pymodule
        else:
            self.old_name = self.old_pyname.get_object().get_name()
            pymodule = self.old_pyname.get_object().get_module()
        self._check_exceptional_conditions()
        self.source = pymodule.get_resource()
        self.tools = _MoveTools(
            self.project, self.source, self.old_pyname, self.old_name
        )
        self.import_tools = self.tools.import_tools

    def _import_filter(self, stmt):
        module_name = libutils.modname(self.source)

        if isinstance(stmt.import_info, importutils.NormalImport):
            # Affect any statement that imports the source module
            return any(
                module_name == name
                for name, alias in stmt.import_info.names_and_aliases
            )
        elif isinstance(stmt.import_info, importutils.FromImport):
            # Affect statements importing from the source package
            if "." in module_name:
                package_name, basename = module_name.rsplit(".", 1)
                if stmt.import_info.module_name == package_name and any(
                    basename == name
                    for name, alias in stmt.import_info.names_and_aliases
                ):
                    return True
            return stmt.import_info.module_name == module_name
        return False

    def _check_exceptional_conditions(self):
        if self._is_variable(self.old_pyname):
            pymodule = self.old_pyname.get_definition_location()[0]
            try:
                pymodule.get_scope().get_name(self.old_name)
            except exceptions.NameNotFoundError:
                self._raise_refactoring_error()
        elif not (
            isinstance(self.old_pyname.get_object(), pyobjects.PyDefinedObject)
            and self._is_global(self.old_pyname.get_object())
        ):
            self._raise_refactoring_error()

    def _raise_refactoring_error(self):
        raise exceptions.RefactoringError(
            "Move refactoring should be performed on a global class, function "
            "or variable."
        )

    def _is_global(self, pyobject):
        return pyobject.get_scope().parent == pyobject.get_module().get_scope()

    def _is_variable(self, pyname):
        return isinstance(pyname, pynames.AssignedName)

    def get_changes(
        self,
        dest: Union[None, str, resources.Resource],
        resources=None,
        task_handle=taskhandle.DEFAULT_TASK_HANDLE,
    ):
        """Return the changes needed for this refactoring

        Parameters:

        - `dest`: the Resource or dotted moddule name of the move destination
        - `resources` can be a list of `rope.base.resources.File` to
          apply this refactoring on.  If `None`, the restructuring
          will be applied to all python files.

        """
        if isinstance(dest, str):
            dest = self.project.find_module(dest)
        if resources is None:
            resources = self.project.get_python_files()
        if dest is None or not dest.exists():
            raise exceptions.RefactoringError("Move destination does not exist.")
        if dest.is_folder() and dest.has_child("__init__.py"):
            dest = dest.get_child("__init__.py")
        if dest.is_folder():
            raise exceptions.RefactoringError(
                "Move destination for non-modules should not be folders."
            )
        if self.source == dest:
            raise exceptions.RefactoringError(
                "Moving global elements to the same module."
            )
        return self._calculate_changes(dest, resources, task_handle)

    def _calculate_changes(self, dest, resources, task_handle):
        changes = ChangeSet("Moving global <%s>" % self.old_name)
        job_set = task_handle.create_jobset("Collecting Changes", len(resources))
        for file_ in resources:
            job_set.started_job(file_.path)
            if file_ == self.source:
                changes.add_change(self._source_module_changes(dest))
            elif file_ == dest:
                changes.add_change(self._dest_module_changes(dest))
            elif self.tools.occurs_in_module(resource=file_):
                pymodule = self.project.get_pymodule(file_)
                # Changing occurrences
                placeholder = "__rope_renaming_%s_" % self.old_name
                source = self.tools.rename_in_module(placeholder, resource=file_)
                should_import = source is not None
                # Removing out of date imports
                pymodule = self.tools.new_pymodule(pymodule, source)
                source = self.import_tools.organize_imports(
                    pymodule, sort=False, import_filter=self._import_filter
                )
                # Adding new import
                if should_import:
                    pymodule = self.tools.new_pymodule(pymodule, source)
                    source, imported = importutils.add_import(
                        self.project, pymodule, self._new_modname(dest), self.old_name
                    )
                    source = source.replace(placeholder, imported)
                source = self.tools.new_source(pymodule, source)
                if source != file_.read():
                    changes.add_change(ChangeContents(file_, source))
            job_set.finished_job()
        return changes

    def _source_module_changes(self, dest):
        placeholder = "__rope_moving_%s_" % self.old_name
        handle = _ChangeMoveOccurrencesHandle(placeholder)
        occurrence_finder = occurrences.create_finder(
            self.project, self.old_name, self.old_pyname
        )
        start, end = self._get_moving_region()
        renamer = ModuleSkipRenamer(occurrence_finder, self.source, handle, start, end)
        source = renamer.get_changed_module()
        pymodule = libutils.get_string_module(self.project, source, self.source)
        source = self.import_tools.organize_imports(pymodule, sort=False)
        if handle.occurred:
            pymodule = libutils.get_string_module(self.project, source, self.source)
            # Adding new import
            source, imported = importutils.add_import(
                self.project, pymodule, self._new_modname(dest), self.old_name
            )
            source = source.replace(placeholder, imported)
        return ChangeContents(self.source, source)

    def _new_modname(self, dest):
        return libutils.modname(dest)

    def _dest_module_changes(self, dest):
        # Changing occurrences
        pymodule = self.project.get_pymodule(dest)
        source = self.tools.rename_in_module(self.old_name, pymodule)
        pymodule = self.tools.new_pymodule(pymodule, source)

        moving, imports = self._get_moving_element_with_imports()
        pymodule, has_changed = self._add_imports2(pymodule, imports)

        module_with_imports = self.import_tools.module_imports(pymodule)
        source = pymodule.source_code
        lineno = 0
        if module_with_imports.imports:
            lineno = module_with_imports.imports[-1].end_line - 1
        else:
            while lineno < pymodule.lines.length() and pymodule.lines.get_line(
                lineno + 1
            ).lstrip().startswith("#"):
                lineno += 1
        if lineno > 0:
            cut = pymodule.lines.get_line_end(lineno) + 1
            result = source[:cut] + "\n\n" + moving + source[cut:]
        else:
            result = moving + source

        # Organizing imports
        source = result
        pymodule = libutils.get_string_module(self.project, source, dest)
        source = self.import_tools.organize_imports(pymodule, sort=False, unused=False)
        # Remove unused imports of the old module
        pymodule = libutils.get_string_module(self.project, source, dest)
        source = self.import_tools.organize_imports(
            pymodule,
            sort=False,
            selfs=False,
            unused=True,
            import_filter=self._import_filter,
        )
        return ChangeContents(dest, source)

    def _get_moving_element_with_imports(self):
        return moving_code_with_imports(
            self.project, self.source, self._get_moving_element()
        )

    def _get_module_with_imports(self, source_code, resource):
        pymodule = libutils.get_string_module(self.project, source_code, resource)
        return self.import_tools.module_imports(pymodule)

    def _get_moving_element(self):
        start, end = self._get_moving_region()
        moving = self.source.read()[start:end]
        return moving.rstrip() + "\n"

    def _get_moving_region(self):
        pymodule = self.project.get_pymodule(self.source)
        lines = pymodule.lines
        if self._is_variable(self.old_pyname):
            logical_lines = pymodule.logical_lines
            lineno = logical_lines.logical_line_in(
                self.old_pyname.get_definition_location()[1]
            )[0]
            start = lines.get_line_start(lineno)
            end_line = logical_lines.logical_line_in(lineno)[1]
        else:
            node = self.old_pyname.get_object().ast_node
            start = lines.get_line_start(
                get_first_decorator_or_function_start_line(node)
            )
            end_line = self.old_pyname.get_object().get_scope().get_end()

        # Include comment lines before the definition
        start_line = lines.get_line_number(start)
        while start_line > 1 and lines.get_line(start_line - 1).startswith("#"):
            start_line -= 1
        start = lines.get_line_start(start_line)

        while end_line < lines.length() and lines.get_line(end_line + 1).strip() == "":
            end_line += 1
        end = min(lines.get_line_end(end_line) + 1, len(pymodule.source_code))
        return start, end

    def _add_imports2(self, pymodule, new_imports):
        source = self.tools.add_imports(pymodule, new_imports)
        if source is None:
            return pymodule, False
        else:
            resource = pymodule.get_resource()
            pymodule = libutils.get_string_module(self.project, source, resource)
            return pymodule, True


class MoveModule:
    """For moving modules and packages"""

    def __init__(self, project, resource):
        self.project = project
        if not resource.is_folder() and resource.name == "__init__.py":
            resource = resource.parent
        if resource.is_folder() and not resource.has_child("__init__.py"):
            raise exceptions.RefactoringError("Cannot move non-package folder.")
        dummy_pymodule = libutils.get_string_module(self.project, "")
        self.old_pyname = pynames.ImportedModule(dummy_pymodule, resource=resource)
        self.source = self.old_pyname.get_object().get_resource()
        if self.source.is_folder():
            self.old_name = self.source.name
        else:
            self.old_name = self.source.name[:-3]
        self.tools = _MoveTools(
            self.project, self.source, self.old_pyname, self.old_name
        )
        self.import_tools = self.tools.import_tools

    def get_changes(
        self, dest, resources=None, task_handle=taskhandle.DEFAULT_TASK_HANDLE
    ):
        if resources is None:
            resources = self.project.get_python_files()
        if dest is None or not dest.is_folder():
            raise exceptions.RefactoringError(
                "Move destination for modules should be packages."
            )
        return self._calculate_changes(dest, resources, task_handle)

    def _calculate_changes(self, dest, resources, task_handle):
        changes = ChangeSet("Moving module <%s>" % self.old_name)
        job_set = task_handle.create_jobset("Collecting changes", len(resources))
        for module in resources:
            job_set.started_job(module.path)
            if module == self.source:
                self._change_moving_module(changes, dest)
            else:
                source = self._change_occurrences_in_module(dest, resource=module)
                if source is not None:
                    changes.add_change(ChangeContents(module, source))
            job_set.finished_job()
        if self.project == self.source.project:
            changes.add_change(MoveResource(self.source, dest.path))
        return changes

    def _new_modname(self, dest):
        destname = libutils.modname(dest)
        if destname:
            return destname + "." + self.old_name
        return self.old_name

    def _new_import(self, dest):
        return importutils.NormalImport([(self._new_modname(dest), None)])

    def _change_moving_module(self, changes, dest):
        if not self.source.is_folder():
            pymodule = self.project.get_pymodule(self.source)
            source = self.import_tools.relatives_to_absolutes(pymodule)
            pymodule = self.tools.new_pymodule(pymodule, source)
            source = self._change_occurrences_in_module(dest, pymodule)
            source = self.tools.new_source(pymodule, source)
            if source != self.source.read():
                changes.add_change(ChangeContents(self.source, source))

    def _change_occurrences_in_module(self, dest, pymodule=None, resource=None):
        if not self.tools.occurs_in_module(pymodule=pymodule, resource=resource):
            return
        if pymodule is None:
            pymodule = self.project.get_pymodule(resource)
        new_name = self._new_modname(dest)
        module_imports = importutils.get_module_imports(self.project, pymodule)
        changed = False
        source = None
        if libutils.modname(dest):
            changed = self._change_import_statements(dest, new_name, module_imports)
            if changed:
                source = module_imports.get_changed_source()
                source = self.tools.new_source(pymodule, source)
                pymodule = self.tools.new_pymodule(pymodule, source)

        new_import = self._new_import(dest)
        source = self.tools.rename_in_module(
            new_name,
            imports=True,
            pymodule=pymodule,
            resource=resource if not changed else None,
        )
        should_import = self.tools.occurs_in_module(
            pymodule=pymodule, resource=resource, imports=False
        )
        pymodule = self.tools.new_pymodule(pymodule, source)
        source = self.tools.remove_old_imports(pymodule)
        if should_import:
            pymodule = self.tools.new_pymodule(pymodule, source)
            source = self.tools.add_imports(pymodule, [new_import])
        source = self.tools.new_source(pymodule, source)
        if source is not None and source != pymodule.resource.read():
            return source
        return None

    def _change_import_statements(self, dest, new_name, module_imports):
        moving_module = self.source
        parent_module = moving_module.parent

        changed = False
        for import_stmt in module_imports.imports:
            if not any(
                name_and_alias[0] == self.old_name
                for name_and_alias in import_stmt.import_info.names_and_aliases
            ) and not any(
                name_and_alias[0] == libutils.modname(self.source)
                for name_and_alias in import_stmt.import_info.names_and_aliases
            ):
                continue

            # Case 1: Look for normal imports of the moving module.
            if isinstance(import_stmt.import_info, importutils.NormalImport):
                continue

            # Case 2: The moving module is from-imported.
            changed = (
                self._handle_moving_in_from_import_stmt(
                    dest, import_stmt, module_imports, parent_module
                )
                or changed
            )

            # Case 3: Names are imported from the moving module.
            context = importutils.importinfo.ImportContext(self.project, None)
            if (
                not import_stmt.import_info.is_empty()
                and import_stmt.import_info.get_imported_resource(context)
                == moving_module
            ):
                import_stmt.import_info = importutils.FromImport(
                    new_name,
                    import_stmt.import_info.level,
                    import_stmt.import_info.names_and_aliases,
                )
                changed = True

        return changed

    def _handle_moving_in_from_import_stmt(
        self, dest, import_stmt, module_imports, parent_module
    ):
        changed = False
        context = importutils.importinfo.ImportContext(self.project, None)
        if import_stmt.import_info.get_imported_resource(context) == parent_module:
            imports = import_stmt.import_info.names_and_aliases
            new_imports = []
            for name, alias in imports:
                # The moving module was imported.
                if name == self.old_name:
                    changed = True
                    new_import = importutils.FromImport(
                        libutils.modname(dest), 0, [(self.old_name, alias)]
                    )
                    module_imports.add_import(new_import)
                else:
                    new_imports.append((name, alias))

            # Update the imports if the imported names were changed.
            if new_imports != imports:
                changed = True
                if new_imports:
                    import_stmt.import_info = importutils.FromImport(
                        import_stmt.import_info.module_name,
                        import_stmt.import_info.level,
                        new_imports,
                    )
                else:
                    import_stmt.empty_import()
        return changed


class _ChangeMoveOccurrencesHandle:
    def __init__(self, new_name):
        self.new_name = new_name
        self.occurred = False

    def occurred_inside_skip(self, change_collector, occurrence):
        pass

    def occurred_outside_skip(self, change_collector, occurrence):
        start, end = occurrence.get_primary_range()
        change_collector.add_change(start, end, self.new_name)
        self.occurred = True


class _MoveTools:
    def __init__(self, project, source, pyname, old_name):
        self.project = project
        self.source = source
        self.old_pyname = pyname
        self.old_name = old_name
        self.import_tools = importutils.ImportTools(self.project)

    def remove_old_imports(self, pymodule):
        old_source = pymodule.source_code
        module_with_imports = self.import_tools.module_imports(pymodule)

        class CanSelect:
            changed = False
            old_name = self.old_name
            old_pyname = self.old_pyname

            def __call__(self, name):
                try:
                    if (
                        name == self.old_name
                        and pymodule[name].get_object() == self.old_pyname.get_object()
                    ):
                        self.changed = True
                        return False
                except exceptions.AttributeNotFoundError:
                    pass
                return True

        can_select = CanSelect()
        module_with_imports.filter_names(can_select)
        new_source = module_with_imports.get_changed_source()
        if old_source != new_source:
            return new_source

    def rename_in_module(self, new_name, pymodule=None, imports=False, resource=None):
        occurrence_finder = self._create_finder(imports)
        source = rename.rename_in_module(
            occurrence_finder,
            new_name,
            replace_primary=True,
            pymodule=pymodule,
            resource=resource,
        )
        return source

    def occurs_in_module(self, pymodule=None, resource=None, imports=True):
        finder = self._create_finder(imports)
        occurrences = finder.find_occurrences(pymodule=pymodule, resource=resource)
        sentinel = object()
        return next(occurrences, sentinel) is not sentinel

    def _create_finder(self, imports):
        return occurrences.create_finder(
            self.project,
            self.old_name,
            self.old_pyname,
            imports=imports,
            keywords=False,
        )

    def new_pymodule(self, pymodule, source):
        if source is not None:
            return libutils.get_string_module(
                self.project, source, pymodule.get_resource()
            )
        return pymodule

    def new_source(self, pymodule, source):
        if source is None:
            return pymodule.source_code
        return source

    def add_imports(self, pymodule, new_imports):
        return _add_imports_to_module(self.import_tools, pymodule, new_imports)


def _add_imports_to_module(import_tools, pymodule, new_imports):
    module_with_imports = import_tools.module_imports(pymodule)
    for new_import in new_imports:
        module_with_imports.add_import(new_import)
    return module_with_imports.get_changed_source()


def moving_code_with_imports(project, resource, source):
    import_tools = importutils.ImportTools(project)
    pymodule = libutils.get_string_module(project, source, resource)

    # Strip comment prefix, if any. These need to stay before the moving
    # section, but imports would be added between them.
    lines = codeanalyze.SourceLinesAdapter(source)
    start = 1
    while start < lines.length() and lines.get_line(start).startswith("#"):
        start += 1
    moving_prefix = source[: lines.get_line_start(start)]
    pymodule = libutils.get_string_module(
        project, source[lines.get_line_start(start) :], resource
    )

    origin = project.get_pymodule(resource)

    imports = [stmt.import_info for stmt in import_tools.module_imports(origin).imports]

    back_names = [name for name in origin if name not in pymodule]
    imports.append(import_tools.get_from_import(resource, back_names))

    source = _add_imports_to_module(import_tools, pymodule, imports)
    pymodule = libutils.get_string_module(project, source, resource)

    source = import_tools.relatives_to_absolutes(pymodule)
    pymodule = libutils.get_string_module(project, source, resource)
    source = import_tools.organize_imports(pymodule, selfs=False)
    pymodule = libutils.get_string_module(project, source, resource)

    # extracting imports after changes
    module_imports = import_tools.module_imports(pymodule)
    imports = [import_stmt.import_info for import_stmt in module_imports.imports]
    start = 1
    if module_imports.imports:
        start = module_imports.imports[-1].end_line
    lines = codeanalyze.SourceLinesAdapter(source)
    while start < lines.length() and not lines.get_line(start).strip():
        start += 1

    # Reinsert the prefix which was removed at the beginning
    moving = moving_prefix + source[lines.get_line_start(start) :]
    return moving, imports


class ModuleSkipRenamerHandle:
    def occurred_outside_skip(self, change_collector, occurrence):
        pass

    def occurred_inside_skip(self, change_collector, occurrence):
        pass


class ModuleSkipRenamer:
    """Rename occurrences in a module

    This class can be used when you want to treat a region in a file
    separately from other parts when renaming.

    """

    def __init__(
        self,
        occurrence_finder,
        resource,
        handle=None,
        skip_start=0,
        skip_end=0,
        replacement="",
    ):
        """Constructor

        if replacement is `None` the region is not changed.  Otherwise
        it is replaced with `replacement`.

        """
        self.occurrence_finder = occurrence_finder
        self.resource = resource
        self.skip_start = skip_start
        self.skip_end = skip_end
        self.replacement = replacement
        self.handle = handle
        if self.handle is None:
            self.handle = ModuleSkipRenamerHandle()

    def get_changed_module(self):
        source = self.resource.read()
        change_collector = codeanalyze.ChangeCollector(source)
        if self.replacement is not None:
            change_collector.add_change(
                self.skip_start, self.skip_end, self.replacement
            )
        for occurrence in self.occurrence_finder.find_occurrences(self.resource):
            start, end = occurrence.get_primary_range()
            if self.skip_start <= start < self.skip_end:
                self.handle.occurred_inside_skip(change_collector, occurrence)
            else:
                self.handle.occurred_outside_skip(change_collector, occurrence)
        result = change_collector.get_changed()
        if result is not None and result != source:
            return result



from __future__ import annotations

from typing import TYPE_CHECKING

from rope.base import (
    change,
    codeanalyze,
    evaluate,
    exceptions,
    libutils,
    pynames,
    pyobjects,
    worder,
)
from rope.refactor import functionutils, importutils, sourceutils, suites

if TYPE_CHECKING:
    from typing import Literal, Optional

    from rope.base.project import Project
    from rope.base.resources import Resource

    GenerateKind = Literal[
        "variable",
        "function",
        "class",
        "module",
        "package",
    ]


def create_generate(
    kind: GenerateKind,
    project: Project,
    resource: Resource,
    offset: int,
    goal_resource: Optional[Resource] = None,
):

    """A factory for creating `Generate` objects

    Used in https://github.com/python-rope/ropemode but not in Rope itself.
    """
    d = {
        "class": GenerateClass,
        "function": GenerateFunction,
        "module": GenerateModule,
        "package": GeneratePackage,
        "variable": GenerateVariable,
    }
    generate = d[kind]
    return generate(project, resource, offset, goal_resource=goal_resource)


def create_module(project, name, sourcefolder=None):
    """Creates a module and returns a `rope.base.resources.File`"""
    if sourcefolder is None:
        sourcefolder = project.root
    packages = name.split(".")
    parent = sourcefolder
    for package in packages[:-1]:
        parent = parent.get_child(package)
    return parent.create_file(packages[-1] + ".py")


def create_package(project, name, sourcefolder=None):
    """Creates a package and returns a `rope.base.resources.Folder`"""
    if sourcefolder is None:
        sourcefolder = project.root
    packages = name.split(".")
    parent = sourcefolder
    for package in packages[:-1]:
        parent = parent.get_child(package)
    made_packages = parent.create_folder(packages[-1])
    made_packages.create_file("__init__.py")
    return made_packages


class _Generate:
    def __init__(self, project, resource, offset, goal_resource=None):
        self.project = project
        self.resource = resource
        self.goal_resource = goal_resource
        self.info = self._generate_info(project, resource, offset)
        self.name = self.info.get_name()
        self._check_exceptional_conditions()

    def _generate_info(self, project, resource, offset):
        return _GenerationInfo(project.pycore, resource, offset, self.goal_resource)

    def _check_exceptional_conditions(self):
        if self.info.element_already_exists():
            raise exceptions.RefactoringError(
                "Element <%s> already exists." % self.name
            )
        if not self.info.primary_is_found():
            raise exceptions.RefactoringError(
                "Cannot determine the scope <%s> should be defined in." % self.name
            )

    def get_changes(self):
        changes = change.ChangeSet(f"Generate {self._get_element_kind()} <{self.name}>")
        indents = self.info.get_scope_indents()
        blanks = self.info.get_blank_lines()
        base_definition = sourceutils.fix_indentation(self._get_element(), indents)
        definition = "\n" * blanks[0] + base_definition + "\n" * blanks[1]

        resource = self.info.get_insertion_resource()
        start, end = self.info.get_insertion_offsets()

        collector = codeanalyze.ChangeCollector(resource.read())
        collector.add_change(start, end, definition)
        changes.add_change(change.ChangeContents(resource, collector.get_changed()))
        if self.goal_resource:
            relative_import = _add_relative_import_to_module(
                self.project, self.resource, self.goal_resource, self.name
            )
            changes.add_change(relative_import)
        return changes

    def get_location(self):
        return (self.info.get_insertion_resource(), self.info.get_insertion_lineno())

    def _get_element_kind(self):
        raise NotImplementedError()

    def _get_element(self):
        raise NotImplementedError()


class GenerateFunction(_Generate):
    def _generate_info(self, project, resource, offset):
        return _FunctionGenerationInfo(project.pycore, resource, offset)

    def _get_element(self):
        decorator = ""
        args = []
        if self.info.is_static_method():
            decorator = "@staticmethod\n"
        if (
            self.info.is_method()
            or self.info.is_constructor()
            or self.info.is_instance()
        ):
            args.append("self")
        args.extend(self.info.get_passed_args())
        definition = "{}def {}({}):\n    pass\n".format(
            decorator,
            self.name,
            ", ".join(args),
        )
        return definition

    def _get_element_kind(self):
        return "Function"


class GenerateVariable(_Generate):
    def _get_element(self):
        return "%s = None\n" % self.name

    def _get_element_kind(self):
        return "Variable"


class GenerateClass(_Generate):
    def _get_element(self):
        return "class %s(object):\n    pass\n" % self.name

    def _get_element_kind(self):
        return "Class"


class GenerateModule(_Generate):
    def get_changes(self):
        package = self.info.get_package()
        changes = change.ChangeSet("Generate Module <%s>" % self.name)
        new_resource = self.project.get_file(f"{package.path}/{self.name}.py")
        if new_resource.exists():
            raise exceptions.RefactoringError(
                "Module <%s> already exists" % new_resource.path
            )
        changes.add_change(change.CreateResource(new_resource))
        changes.add_change(
            _add_import_to_module(self.project, self.resource, new_resource)
        )
        return changes

    def get_location(self):
        package = self.info.get_package()
        return (package.get_child("%s.py" % self.name), 1)


class GeneratePackage(_Generate):
    def get_changes(self):
        package = self.info.get_package()
        changes = change.ChangeSet("Generate Package <%s>" % self.name)
        new_resource = self.project.get_folder(f"{package.path}/{self.name}")
        if new_resource.exists():
            raise exceptions.RefactoringError(
                "Package <%s> already exists" % new_resource.path
            )
        changes.add_change(change.CreateResource(new_resource))
        changes.add_change(
            _add_import_to_module(self.project, self.resource, new_resource)
        )
        child = self.project.get_folder(package.path + "/" + self.name)
        changes.add_change(change.CreateFile(child, "__init__.py"))
        return changes

    def get_location(self):
        package = self.info.get_package()
        child = package.get_child(self.name)
        return (child.get_child("__init__.py"), 1)


def _add_import_to_module(project, resource, imported):
    pymodule = project.get_pymodule(resource)
    import_tools = importutils.ImportTools(project)
    module_imports = import_tools.module_imports(pymodule)
    module_name = libutils.modname(imported)
    new_import = importutils.NormalImport(((module_name, None),))
    module_imports.add_import(new_import)
    return change.ChangeContents(resource, module_imports.get_changed_source())


def _add_relative_import_to_module(project, resource, imported, name):
    pymodule = project.get_pymodule(resource)
    import_tools = importutils.ImportTools(project)
    module_imports = import_tools.module_imports(pymodule)
    new_import = import_tools.get_from_import(imported, name)
    module_imports.add_import(new_import)
    return change.ChangeContents(resource, module_imports.get_changed_source())


class _GenerationInfo:
    def __init__(self, pycore, resource, offset, goal_resource=None):
        self.pycore = pycore
        self.resource = resource
        self.offset = offset
        self.goal_resource = goal_resource
        self.source_pymodule = self.pycore.project.get_pymodule(resource)
        finder = evaluate.ScopeNameFinder(self.source_pymodule)
        self.primary, self.pyname = finder.get_primary_and_pyname_at(offset)
        self._init_fields()

    def _init_fields(self):
        self.source_scope = self._get_source_scope()
        self.goal_scope = self._get_goal_scope()
        self.goal_pymodule = self._get_goal_module(self.goal_scope)

    def _get_goal_scope(self):
        if self.primary is None:
            if self.goal_resource:
                return self.pycore.project.get_pymodule(self.goal_resource).get_scope()
            else:
                return self._get_source_scope()
        pyobject = self.primary.get_object()
        if isinstance(pyobject, pyobjects.PyDefinedObject):
            return pyobject.get_scope()
        elif isinstance(pyobject.get_type(), pyobjects.PyClass):
            return pyobject.get_type().get_scope()

    def _get_goal_module(self, scope):
        if scope is None:
            return
        while scope.parent is not None:
            scope = scope.parent
        return scope.pyobject

    def _get_source_scope(self):
        module_scope = self.source_pymodule.get_scope()
        lineno = self.source_pymodule.lines.get_line_number(self.offset)
        return module_scope.get_inner_scope_for_line(lineno)

    def get_insertion_lineno(self):
        lines = self.goal_pymodule.lines
        if self.goal_scope == self.source_scope:
            line_finder = self.goal_pymodule.logical_lines
            lineno = lines.get_line_number(self.offset)
            lineno = line_finder.logical_line_in(lineno)[0]
            root = suites.ast_suite_tree(self.goal_scope.pyobject.get_ast())
            suite = root.find_suite(lineno)
            indents = sourceutils.get_indents(lines, lineno)
            while self.get_scope_indents() < indents:
                lineno = suite.get_start()
                indents = sourceutils.get_indents(lines, lineno)
                suite = suite.parent
            return lineno
        else:
            return min(self.goal_scope.get_end() + 1, lines.length())

    def get_insertion_resource(self):
        return self.goal_pymodule.get_resource()

    def get_insertion_offsets(self):
        if self.goal_scope.get_kind() == "Class":
            start, end = sourceutils.get_body_region(self.goal_scope.pyobject)
            if self.goal_pymodule.source_code[start:end].strip() == "pass":
                return start, end
        lines = self.goal_pymodule.lines
        start = lines.get_line_start(self.get_insertion_lineno())
        return (start, start)

    def get_scope_indents(self):
        if self.goal_scope.get_kind() == "Module":
            return 0
        return (
            sourceutils.get_indents(
                self.goal_pymodule.lines, self.goal_scope.get_start()
            )
            + 4
        )

    def get_blank_lines(self):
        if self.goal_scope.get_kind() == "Module":
            base_blanks = 2
            if self.goal_pymodule.source_code.strip() == "":
                base_blanks = 0
        if self.goal_scope.get_kind() == "Class":
            base_blanks = 1
        if self.goal_scope.get_kind() == "Function":
            base_blanks = 0
        if self.goal_scope == self.source_scope:
            return (0, base_blanks)
        return (base_blanks, 0)

    def get_package(self):
        primary = self.primary
        if self.primary is None:
            return self.pycore.project.get_source_folders()[0]
        if isinstance(primary.get_object(), pyobjects.PyPackage):
            return primary.get_object().get_resource()
        raise exceptions.RefactoringError(
            "A module/package can be only created in a package."
        )

    def primary_is_found(self):
        return self.goal_scope is not None

    def element_already_exists(self):
        if self.pyname is None or isinstance(self.pyname, pynames.UnboundName):
            return False
        return self.get_name() in self.goal_scope.get_defined_names()

    def get_name(self):
        return worder.get_name_at(self.resource, self.offset)


class _FunctionGenerationInfo(_GenerationInfo):
    def _get_goal_scope(self):
        if self.is_constructor():
            return self.pyname.get_object().get_scope()
        if self.is_instance():
            return self.pyname.get_object().get_type().get_scope()
        if self.primary is None:
            return self._get_source_scope()
        pyobject = self.primary.get_object()
        if isinstance(pyobject, pyobjects.PyDefinedObject):
            return pyobject.get_scope()
        elif isinstance(pyobject.get_type(), pyobjects.PyClass):
            return pyobject.get_type().get_scope()

    def element_already_exists(self):
        if self.pyname is None or isinstance(self.pyname, pynames.UnboundName):
            return False
        return self.get_name() in self.goal_scope.get_defined_names()

    def is_static_method(self):
        return self.primary is not None and isinstance(
            self.primary.get_object(), pyobjects.PyClass
        )

    def is_method(self):
        return self.primary is not None and isinstance(
            self.primary.get_object().get_type(), pyobjects.PyClass
        )

    def is_constructor(self):
        return self.pyname is not None and isinstance(
            self.pyname.get_object(), pyobjects.PyClass
        )

    def is_instance(self):
        if self.pyname is None:
            return False
        pyobject = self.pyname.get_object()
        return isinstance(pyobject.get_type(), pyobjects.PyClass)

    def get_name(self):
        if self.is_constructor():
            return "__init__"
        if self.is_instance():
            return "__call__"
        return worder.get_name_at(self.resource, self.offset)

    def get_passed_args(self):
        result = []
        source = self.source_pymodule.source_code
        finder = worder.Worder(source)
        if finder.is_a_function_being_called(self.offset):
            start, end = finder.get_primary_range(self.offset)
            parens_start, parens_end = finder.get_word_parens_range(end - 1)
            call = source[start:parens_end]
            parser = functionutils._FunctionParser(call, False)
            args, keywords = parser.get_parameters()
            for arg in args:
                if self._is_id(arg):
                    result.append(arg)
                else:
                    result.append("arg%d" % len(result))
            for name, value in keywords:
                result.append(name)
        return result

    def _is_id(self, arg):
        def id_or_underline(c):
            return c.isalpha() or c == "_"

        for c in arg:
            if not id_or_underline(c) and not c.isdigit():
                return False
        return id_or_underline(arg[0])



import unittest
from textwrap import dedent

from rope.base import exceptions
from rope.contrib.findit import find_definition, find_implementations, find_occurrences
from ropetest import testutils


class FindItTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_finding_occurrences(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write("a_var = 1\n")
        result = find_occurrences(self.project, mod, 1)
        self.assertEqual(mod, result[0].resource)
        self.assertEqual(0, result[0].offset)
        self.assertEqual(False, result[0].unsure)

    def test_finding_occurrences_in_more_than_one_module(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("a_var = 1\n")
        mod2.write(dedent("""\
            import mod1
            my_var = mod1.a_var"""))
        result = find_occurrences(self.project, mod1, 1)
        self.assertEqual(2, len(result))
        modules = (result[0].resource, result[1].resource)
        self.assertTrue(mod1 in modules and mod2 in modules)

    def test_finding_occurrences_matching_when_unsure(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class C(object):
                def a_func(self):
                    pass
            def f(arg):
                arg.a_func()
        """))
        result = find_occurrences(
            self.project, mod1, mod1.read().index("a_func"), unsure=True
        )
        self.assertEqual(2, len(result))

    def test_find_occurrences_resources_parameter(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("a_var = 1\n")
        mod2.write(dedent("""\
            import mod1
            my_var = mod1.a_var"""))
        result = find_occurrences(self.project, mod1, 1, resources=[mod1])
        self.assertEqual(1, len(result))
        self.assertEqual((mod1, 0), (result[0].resource, result[0].offset))

    def test_find_occurrences_and_class_hierarchies(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class A(object):
                def f():
                    pass
            class B(A):
                def f():
                    pass
        """))
        offset = mod1.read().rindex("f")
        result1 = find_occurrences(self.project, mod1, offset)
        result2 = find_occurrences(self.project, mod1, offset, in_hierarchy=True)
        self.assertEqual(1, len(result1))
        self.assertEqual(2, len(result2))

    def test_trivial_find_implementations(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class A(object):
                def f(self):
                    pass
        """))
        offset = mod1.read().rindex("f(")
        result = find_implementations(self.project, mod1, offset)
        self.assertEqual([], result)

    def test_find_implementations_and_not_returning_parents(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class A(object):
                def f(self):
                    pass
            class B(A):
                def f(self):
                    pass
        """))
        offset = mod1.read().rindex("f(")
        result = find_implementations(self.project, mod1, offset)
        self.assertEqual([], result)

    def test_find_implementations_real_implementation(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            class A(object):
                def f(self):
                    pass
            class B(A):
                def f(self):
                    pass
        """))
        offset = mod1.read().index("f(")
        result = find_implementations(self.project, mod1, offset)
        self.assertEqual(1, len(result))
        self.assertEqual(mod1.read().rindex("f("), result[0].offset)

    def test_find_implementations_real_implementation_simple(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("class A(object):\n    pass\n")
        offset = mod1.read().index("A")
        with self.assertRaises(exceptions.BadIdentifierError):
            find_implementations(self.project, mod1, offset)

    def test_trivial_find_definition(self):
        code = dedent("""\
            def a_func():
                pass
            a_func()""")
        result = find_definition(self.project, code, code.rindex("a_func"))
        start = code.index("a_func")
        self.assertEqual(start, result.offset)
        self.assertEqual(None, result.resource)
        self.assertEqual(1, result.lineno)
        self.assertEqual((start, start + len("a_func")), result.region)

    def test_find_definition_in_other_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("var = 1\n")
        code = dedent("""\
            import mod1
            print(mod1.var)
        """)
        result = find_definition(self.project, code, code.index("var"))
        self.assertEqual(mod1, result.resource)
        self.assertEqual(0, result.offset)



"""Fix the name of modules

This module is useful when you want to rename many of the modules in
your project.  That can happen specially when you want to change their
naming style.

For instance::

  fixer = FixModuleNames(project)
  changes = fixer.get_changes(fixer=str.lower)
  project.do(changes)

Here it renames all modules and packages to use lower-cased chars.
You can tell it to use any other style by using the ``fixer``
argument.

"""
from rope.base import taskhandle
from rope.contrib import changestack
from rope.refactor import rename


class FixModuleNames:
    def __init__(self, project):
        self.project = project

    def get_changes(self, fixer=str.lower, task_handle=taskhandle.DEFAULT_TASK_HANDLE):
        """Fix module names

        `fixer` is a function that takes and returns a `str`.  Given
        the name of a module, it should return the fixed name.

        """
        stack = changestack.ChangeStack(self.project, "Fixing module names")
        jobset = task_handle.create_jobset(
            "Fixing module names", self._count_fixes(fixer) + 1
        )
        try:
            while True:
                for resource in self._tobe_fixed(fixer):
                    jobset.started_job(resource.path)
                    renamer = rename.Rename(self.project, resource)
                    changes = renamer.get_changes(fixer(self._name(resource)))
                    stack.push(changes)
                    jobset.finished_job()
                    break
                else:
                    break
        finally:
            jobset.started_job("Reverting to original state")
            stack.pop_all()
            jobset.finished_job()
        return stack.merged()

    def _count_fixes(self, fixer):
        return len(list(self._tobe_fixed(fixer)))

    def _tobe_fixed(self, fixer):
        for resource in self.project.get_python_files():
            modname = self._name(resource)
            if modname != fixer(modname):
                yield resource

    def _name(self, resource):
        modname = resource.name.rsplit(".", 1)[0]
        if modname == "__init__":
            modname = resource.parent.name
        return modname
